"""tests"""
import datetime

from pyiem import prism


def test_ij():
    """Can we get valid indices back!"""
    res = prism.find_ij(-98.0, 32)
    assert res[0] == 647

    res = prism.find_ij(98.0, 32)
    assert res[0] is None


def test_tidx():
    """Can we get time indices"""
    valid = datetime.datetime(2017, 9, 1)
    tidx = prism.daily_offset(valid)
    assert tidx == 243



""" Handle some of the fun things that come from the Hurricane Center """

import re

from pyiem.nws.product import TextProduct
from pyiem.exceptions import NHCException
from pyiem.reference import TWEET_CHARS

TITLE = (
    "(POST-TROPICAL CYCLONE|TROPICAL STORM|HURRICANE|"
    "POTENTIAL TROPICAL CYCLONE|TROPICAL CYCLONE|"
    r"TROPICAL DEPRESSION|REMNANTS OF) ([A-Z0-9\- ]*) "
    "(DISCUSSION|INTERMEDIATE ADVISORY|FORECAST/ADVISORY|ADVISORY) "
    r"NUMBER\s+([0-9A-Z]+)"
)


class NHCProduct(TextProduct):
    """
    Represents a NHC
    """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)

    def get_jabbers(self, uri, _uri2=None):
        """ Get the jabber variant of this message """
        myurl = "%s?pid=%s" % (uri, self.get_product_id())

        tokens = re.findall(TITLE, self.unixtext.upper().replace("\n", " "))
        if not tokens:
            if self.source != "KNHC":
                return []
            raise NHCException("Could not parse header from NHC Product!")

        classification = tokens[0][0]
        name = tokens[0][1]
        twitter_name = name.title().replace("-", "_")
        prodtype = tokens[0][2]
        prodnumber = tokens[0][3]
        center = "National Hurricane Center"

        tformat = (
            "%(classification)s #%(storm_name)s "
            "%(btype)s %(num)s issued. %(headline)s "
            "http://go.usa.gov/W3H"
        )
        tdict = {
            "classification": classification.title(),
            "storm_name": twitter_name,
            "num": prodnumber,
            "btype": prodtype,
            "headline": "",
            "url": myurl,
        }

        mess = "%s issues %s #%s for %s %s %s" % (
            center,
            prodtype,
            prodnumber,
            classification,
            name,
            myurl,
        )
        htmlmess = '%s issues <a href="%s">%s #%s</a> for %s %s' % (
            center,
            myurl,
            prodtype,
            prodnumber,
            classification,
            name,
        )

        if self.segments[0].headlines:
            headline = self.segments[0].headlines[0]
            headline = headline.lower().replace(
                name.lower(), "#%s" % (twitter_name,)
            )
            headline = headline[0].upper() + headline[1:] + "."
            if (TWEET_CHARS - len(tformat % tdict)) > len(headline):
                tdict["headline"] = headline
            else:
                headline = headline[: headline.find(",")]
                if (TWEET_CHARS - len(tformat % tdict)) > len(headline):
                    tdict["headline"] = headline

        tweet = tformat % tdict

        return [
            [
                mess.replace("#", "") % tdict,
                htmlmess.replace("#", "") % tdict,
                {
                    "channels": "NHC,%s,%s,%s"
                    % (self.afos[:5], name, self.afos),
                    "product_id": self.get_product_id(),
                    "twitter": " ".join(tweet.split()),
                },
            ]
        ]


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Helper function """
    return NHCProduct(text, utcnow, ugc_provider, nwsli_provider)



"""
 Supports parsing of Storm Prediction Center's MCD and
 parsing of Weather Prediction Center's MPD
"""
import re
from datetime import timezone, timedelta

from shapely.geometry import Polygon as ShapelyPolygon
from pyiem.nws.product import TextProduct
from pyiem.exceptions import MCDException
from pyiem.reference import TWEET_CHARS
from pyiem.util import html_escape, LOG

LATLON = re.compile(r"LAT\.\.\.LON\s+((?:[0-9]{8}\s+)+)")
DISCUSSIONNUM = re.compile(
    r"MESOSCALE (?:PRECIPITATION )?DISCUSSION\s+([0-9]+)", re.IGNORECASE
)
ATTN_WFO = re.compile(
    r"ATTN\.\.\.WFO\.\.\.([\.A-Z]*?)(?:LAT\.\.\.LON|ATTN\.\.\.RFC)"
)
ATTN_RFC = re.compile(r"ATTN\.\.\.RFC\.\.\.([\.A-Z]*)")
WATCH_PROB = re.compile(
    r"PROBABILITY OF WATCH ISSUANCE\s?\.\.\.\s?([0-9]+) PERCENT", re.IGNORECASE
)
VALID_TIME = re.compile(r"VALID\s+([0-9]{6})Z?\s?-\s?([0-9]{6})Z?", re.I)
CONCERNING = re.compile(r"CONCERNING\s?\.\.\.(.*?)\.\.\.", re.I)


class MCDProduct(TextProduct):
    """
    Represents a Storm Prediction Center Mesoscale Convective Discussion
    """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)
        self.geometry = self.parse_geometry()
        self.discussion_num = self.parse_discussion_num()
        self.attn_wfo = self.parse_attn_wfo()
        self.attn_rfc = self.parse_attn_rfc()
        self.areas_affected = self.parse_areas_affected()
        self.concerning = self.parse_concerning()
        self.watch_prob = self.find_watch_probability()
        self.sts, self.ets = self.find_valid_times()
        self.cwsus = []

    def find_valid_times(self):
        """Figure out when this product is valid for"""
        tokens = VALID_TIME.findall(self.unixtext)
        if not tokens:
            self.warnings.append("failed to find VALID...")
            return None, None
        day1 = int(tokens[0][0][:2])
        hour1 = int(tokens[0][0][2:4])
        min1 = int(tokens[0][0][4:])
        day2 = int(tokens[0][1][:2])
        hour2 = int(tokens[0][1][2:4])
        min2 = int(tokens[0][1][4:])
        issue = self.valid.replace(day=day1, hour=hour1, minute=min1)
        expire = self.valid.replace(day=day2, hour=hour2, minute=min2)
        if day1 < self.valid.day and day1 == 1:
            issue = self.valid + timedelta(days=25)
            issue = issue.replace(day=day1, hour=hour1, minute=min1)
        if day2 < self.valid.day and day2 == 1:
            expire = self.valid + timedelta(days=25)
            expire = expire.replace(day=day2, hour=hour2, minute=min2)

        return (
            issue.replace(tzinfo=timezone.utc),
            expire.replace(tzinfo=timezone.utc),
        )

    def find_watch_probability(self):
        """ Find the probability of watch issuance for SPC MCD"""
        tokens = WATCH_PROB.findall(self.unixtext.replace("\n", ""))
        if not tokens:
            return None
        return int(tokens[0])

    def tweet(self):
        """ Return twitter message """
        charsleft = TWEET_CHARS - 24  # default safe 24 for t.co shortening
        if self.afos == "SWOMCD":
            center = "SPC"
        else:
            center = "WPC"
        prob_extra = ""
        if self.watch_prob is not None:
            prob_extra = " [watch prob: %.0f%%]" % (self.watch_prob,)
        concerning_text = ""
        if self.concerning is not None:
            concerning_text = f" concerning {self.concerning}"
        attempt = ("#%s issues %s %s%s%s: %s ") % (
            center,
            self.afos[3:],
            self.discussion_num,
            concerning_text,
            prob_extra,
            self.areas_affected,
        )
        return "%s%s" % (attempt[:charsleft], self.get_url())

    def get_url(self):
        """ Return the URL for SPC's website """
        if self.afos == "SWOMCD":
            return ("https://www.spc.noaa.gov/products/md/%s/md%04i.html") % (
                self.valid.year,
                self.discussion_num,
            )
        return (
            "https://www.wpc.ncep.noaa.gov/metwatch/"
            "metwatch_mpd_multi.php?md=%s&yr=%s"
        ) % (self.discussion_num, self.valid.year)

    def parse_areas_affected(self):
        """ Return the areas affected """
        sections = self.unixtext.split("\n\n")
        for section in sections:
            if section.strip().find("AREAS AFFECTED...") == 0:
                return section[17:].replace("\n", " ")
        return None

    def get_jabbers(self, uri, _uri2=None):
        """Return plain text and html variants for a Jabber msg

        Args:
          uri (str): URL number one needed for constructing the URI
          _uri2 (str): not used, but needed for the over-ride

        Returns:
          (list): List of lists, plain text, html text, xtra dict
        """
        # convert htmlentities
        spcuri = html_escape(self.get_url())
        center = "Storm Prediction Center"
        pextra = ""
        if self.afos == "FFGMPD":
            center = "Weather Prediction Center"
            pextra = "Precipitation "
        prob_extra = " "
        if self.watch_prob is not None:
            prob_extra = " [watch probability: %.0f%%] " % (self.watch_prob,)
        concerning_text = ""
        if self.concerning is not None:
            concerning_text = f" concerning {self.concerning}"
        plain = ("%s issues Mesoscale %sDiscussion #%s%s%s%s") % (
            center,
            pextra,
            self.discussion_num,
            concerning_text,
            prob_extra,
            spcuri,
        )
        html = (
            '<p>%s issues <a href="%s">'
            "Mesoscale %sDiscussion #%s</a>%s%s"
            '(<a href="%s?pid=%s">View text</a>)</p>'
        ) % (
            center,
            spcuri,
            pextra,
            self.discussion_num,
            concerning_text,
            prob_extra,
            uri,
            self.get_product_id(),
        )
        channels = self.get_channels()
        channels.extend(["%s.%s" % (self.afos, w) for w in self.attn_wfo])
        channels.extend(["%s.%s" % (self.afos, w) for w in self.attn_rfc])
        channels.extend(["%s.%s" % (self.afos, w) for w in self.cwsus])
        xtra = dict(
            channels=",".join(channels),
            product_id=self.get_product_id(),
            twitter=self.tweet(),
        )
        return [[plain, html, xtra]]

    def parse_concerning(self):
        """Figure out the concerning text, if it exists."""
        tokens = CONCERNING.findall(self.unixtext.replace("\n", ""))
        if not tokens:
            return None
        return tokens[0]

    def parse_attn_rfc(self):
        """ Figure out which RFCs this product is seeking attention """
        tokens = ATTN_RFC.findall(self.unixtext.replace("\n", ""))
        if not tokens:
            return []
        return re.findall("([A-Z]{5})", tokens[0])

    def parse_attn_wfo(self):
        """ Figure out which WFOs this product is seeking attention """
        tokens = ATTN_WFO.findall(self.unixtext.replace("\n", ""))
        if not tokens:
            raise MCDException("Could not parse attention WFOs")
        return re.findall("([A-Z]{3})", tokens[0])

    def parse_discussion_num(self):
        """ Figure out what discussion number this is """
        tokens = DISCUSSIONNUM.findall(self.unixtext)
        if not tokens:
            raise MCDException("Could not parse discussion number")
        return int(tokens[0])

    def parse_geometry(self):
        """ Find the polygon that's in this MCD product """
        tokens = LATLON.findall(self.unixtext.replace("\n", " "))
        if not tokens:
            raise MCDException("Could not parse LAT...LON geometry")
        pts = []
        for pair in tokens[0].split():
            lat = float(pair[:4]) / 100.0
            lon = 0 - float(pair[4:]) / 100.0
            if lon > -40:
                lon = lon - 100.0
            pts.append((lon, lat))
        return ShapelyPolygon(pts)

    def database_save(self, txn):
        """Save this product to the database"""
        table = "mcd" if self.afos == "SWOMCD" else "mpd"
        # Remove any previous entries
        sql = f"DELETE from {table} where product_id = %s and num = %s"
        txn.execute(sql, (self.get_product_id(), self.discussion_num))
        if txn.rowcount > 0:
            LOG.info(
                "mcd.database_save %s %s removed %s entries",
                self.get_product_id(),
                self.discussion_num,
                txn.rowcount,
            )
        giswkt = "SRID=4326;%s" % (self.geometry.wkt,)
        sql = (
            f"INSERT into {table} (product, product_id, geom, issue, expire, "
            "num, year, watch_confidence, concerning) "
            "values (%s, %s, %s, %s, %s, %s, %s, %s, %s)"
        )
        args = (
            self.text,
            self.get_product_id(),
            giswkt,
            self.sts,
            self.ets,
            self.discussion_num,
            self.valid.year,
            self.find_watch_probability(),
            self.concerning,
        )
        txn.execute(sql, args)


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Helper function """
    return MCDProduct(text, utcnow, ugc_provider, nwsli_provider)



"""tests"""

import pytest
from shapely.geometry import Point, Polygon, LineString
from pyiem import wellknowntext


def test_wkt():
    """ Try the properties function"""
    wkt = "SRID=4326;POINT(-99 43)"
    geom = wellknowntext.convert_well_known_text(wkt)
    assert Point(geom) == Point([-99, 43])

    wkt = """MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)),
        ((20 35, 10 30, 10 10, 30 5, 45 20, 20 35),
        (30 20, 20 15, 20 25, 30 20)))"""
    geom = wellknowntext.convert_well_known_text(wkt)
    assert abs(Polygon(geom[0]).area - 87.5) < 0.1

    wkt = """MULTILINESTRING ((10 10, 20 20, 10 40),
    (40 40, 30 30, 40 20, 30 10))"""
    geom = wellknowntext.convert_well_known_text(wkt)
    assert abs(LineString(geom[0]).length - 36.5) < 0.1

    wkt = """LINESTRING (30 10, 10 30, 40 40)"""
    geom = wellknowntext.convert_well_known_text(wkt)
    assert abs(LineString(geom).length - 59.9) < 0.1

    wkt = """POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))"""
    geom = wellknowntext.convert_well_known_text(wkt)
    assert abs(Polygon(geom[0]).area - 550.0) < 0.1

    wkt = """POLYGON q((30 10, 40 40, 20 40, 10 20, 30 10))q"""
    with pytest.raises(ValueError):
        wellknowntext.convert_well_known_text(wkt)

    wkt = """RARRR q((30 10, 40 40, 20 40, 10 20, 30 10))q"""
    with pytest.raises(ValueError):
        wellknowntext.convert_well_known_text(wkt)

    with pytest.raises(ValueError):
        wellknowntext.convert_well_known_text("")
